# menu.py
#
# This file is part of Wike, a Wikipedia Reader for the GNOME Desktop.
# Copyright 2021 Hugo Olabera <hugolabe@gmail.com>.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gio, Gdk, Gtk

from wike.data import settings


# Popover class for main menu
# Show and manage main menu

@Gtk.Template(resource_path='/com/github/hugolabe/Wike/ui/menu.ui')
class MenuPopover(Gtk.Popover):

  __gtype_name__ = 'MenuPopover'
  
  system_button = Gtk.Template.Child()
  light_button = Gtk.Template.Child()
  sepia_button = Gtk.Template.Child()
  dark_button = Gtk.Template.Child()

  # Load custom css for theme buttons and connect signals

  def __init__(self, window):
    super().__init__()

    self._window = window
    
    gfile = Gio.File.new_for_uri('resource:///com/github/hugolabe/Wike/ui/menu.css')
    css_provider = Gtk.CssProvider()
    css_provider.load_from_file(gfile)
    Gtk.StyleContext.add_provider_for_screen(Gdk.Screen.get_default(), css_provider, Gtk.STYLE_PROVIDER_PRIORITY_USER)
    
    theme = settings.get_int('theme')
    if theme == 0:
      self.light_button.set_active(True)
    elif theme == 1:
      self.dark_button.set_active(True)
    elif theme == 2:
      self.sepia_button.set_active(True)
    else:
      self.system_button.set_active(True)

    self.system_button.connect('toggled', self._system_button_toggled_cb)
    self.light_button.connect('toggled', self._light_button_toggled_cb)
    self.sepia_button.connect('toggled', self._sepia_button_toggled_cb)
    self.dark_button.connect('toggled', self._dark_button_toggled_cb)
  
  # Change to system theme on button toggle

  def _system_button_toggled_cb(self, button):
    if button.get_active():
      app = self._window.get_application()
      theme_action = app.lookup_action('theme_system')
      theme_action.activate()

  # Change to light theme on button toggle
  
  def _light_button_toggled_cb(self, button):
    if button.get_active():
      app = self._window.get_application()
      theme_action = app.lookup_action('theme_light')
      theme_action.activate()

  # Change to sepia theme on button toggle
  
  def _sepia_button_toggled_cb(self, button):
    if button.get_active():
      app = self._window.get_application()
      theme_action = app.lookup_action('theme_sepia')
      theme_action.activate()

  # Change to dark theme on button toggle
  
  def _dark_button_toggled_cb(self, button):
    if button.get_active():
      app = self._window.get_application()
      theme_action = app.lookup_action('theme_dark')
      theme_action.activate()
    



# application.py
#
# This file is part of Wike, a Wikipedia Reader for the GNOME Desktop.
# Copyright 2021 Hugo Olabera <hugolabe@gmail.com>.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import sys

import gi
gi.require_version('Gtk', '3.0')
gi.require_version('Handy', '1')
from gi.repository import Gio, GLib, Gtk, Handy

from wike.data import settings, languages, historic, bookmarks
from wike.prefs import PrefsWindow
from wike.window import Window


# Main application class
# Wike Wikipedia Reader

class Application(Gtk.Application):

  # Initialize app

  def __init__(self):
    super().__init__(application_id='com.github.hugolabe.Wike', flags=Gio.ApplicationFlags.HANDLES_COMMAND_LINE)

    self._window = None
    self._launch_uri = ''

    self.add_main_option('url', b'u', GLib.OptionFlags.NONE, GLib.OptionArg.STRING, 'Wikipedia URL to open', None)

  # Initialize libhandy and set actions

  def do_startup(self):
    Gtk.Application.do_startup(self)
    Handy.init()

    self._style_manager = Handy.StyleManager.get_default()
    if settings.get_int('theme') == 0:
      self._style_manager.set_color_scheme(Handy.ColorScheme.FORCE_LIGHT)
    elif settings.get_int('theme') == 1:
      self._style_manager.set_color_scheme(Handy.ColorScheme.FORCE_DARK)
    elif settings.get_int('theme') == 2:
      self._style_manager.set_color_scheme(Handy.ColorScheme.FORCE_LIGHT)
    elif settings.get_int('theme') == 3:
      self._style_manager.set_color_scheme(Handy.ColorScheme.PREFER_LIGHT)

    action = Gio.SimpleAction.new('prefs', None)
    action.connect('activate', self._prefs_cb)
    self.add_action(action)
    self.set_accels_for_action('app.prefs', ('<Ctrl>E',))

    action = Gio.SimpleAction.new('theme_system', None)
    action.connect('activate', self._theme_system)
    self.add_action(action)

    action = Gio.SimpleAction.new('theme_light', None)
    action.connect('activate', self._theme_light)
    self.add_action(action)

    action = Gio.SimpleAction.new('theme_sepia', None)
    action.connect('activate', self._theme_sepia)
    self.add_action(action)

    action = Gio.SimpleAction.new('theme_dark', None)
    action.connect('activate', self._theme_dark)
    self.add_action(action)

    action = Gio.SimpleAction.new('shortcuts', None)
    action.connect('activate', self._shortcuts_cb)
    self.add_action(action)
    self.set_accels_for_action('app.shortcuts', ('<Ctrl>question',))

    action = Gio.SimpleAction.new('about', None)
    action.connect('activate', self._about_cb)
    self.add_action(action)

    action = Gio.SimpleAction.new('quit', None)
    action.connect('activate', self._quit_cb)
    self.add_action(action)
    self.set_accels_for_action('app.quit', ('<Ctrl>Q',))

  # Manage command line options

  def do_command_line(self, command_line):
    options = command_line.get_options_dict().end().unpack()

    if 'url' in options:
      self._launch_uri = options['url']

    if self._window:
      self._window.new_page(self._launch_uri, None, True)

    self.activate()
    return 0

  # Create main window and connect delete event

  def do_activate(self):
    if not self._window:
      self._window = Window(self, self._launch_uri)
      self._window.connect('delete-event',self._window_delete_cb)

      self._window.show_all()
    else:
      self._window.present()

  # Show Preferences window

  def _prefs_cb(self, action, parameter):
    prefs_window = PrefsWindow()
    prefs_window.set_transient_for(self._window)
    prefs_window.show_all()

  # Set theme system

  def _theme_system(self, action, parameter):
    self._style_manager.set_color_scheme(Handy.ColorScheme.PREFER_LIGHT)
    settings.set_int('theme', 3)

  # Set theme light

  def _theme_light(self, action, parameter):
    self._style_manager.set_color_scheme(Handy.ColorScheme.FORCE_LIGHT)
    settings.set_int('theme', 0)

  # Set theme sepia

  def _theme_sepia(self, action, parameter):
    self._style_manager.set_color_scheme(Handy.ColorScheme.FORCE_LIGHT)
    settings.set_int('theme', 2)

  # Set theme dark

  def _theme_dark(self, action, parameter):
    self._style_manager.set_color_scheme(Handy.ColorScheme.FORCE_DARK)
    settings.set_int('theme', 1)

  # Show Shortcuts window

  def _shortcuts_cb(self, action, parameter):
    builder = Gtk.Builder()
    builder.add_from_resource("/com/github/hugolabe/Wike/ui/shortcuts.ui")
    shortcuts_window = builder.get_object("shortcuts_window")
    shortcuts_window.set_transient_for(self._window)
    shortcuts_window.show()

  # Show About dialog

  def _about_cb(self, action, parameter):
    builder = Gtk.Builder()
    builder.add_from_resource("/com/github/hugolabe/Wike/ui/about.ui")
    about_dialog = builder.get_object("about_dialog")
    about_dialog.set_transient_for(self._window)
    about_dialog.run()
    about_dialog.destroy()

  # On window delete quit app

  def _window_delete_cb(self, window, event):
    window.tabview.disconnect(window.handler_selpage)
    quit_action = self.lookup_action('quit')
    quit_action.activate()

  # Save settings and data and quit app

  def _quit_cb(self, action, parameter):
    if self._window.is_maximized():
      settings.set_boolean('window-max', True)
    else:
      settings.set_boolean('window-max', False)
      size = self._window.get_size()
      settings.set_int('window-width', size[0])
      settings.set_int('window-height', size[1])

    if self._window.page.wikiview.is_local():
      settings.set_string('last-uri', '')
    else:
      settings.set_string('last-uri', self._window.page.wikiview.get_base_uri())

    settings.sync()
    languages.save()
    if not settings.get_boolean('keep-historic'):
      historic.clear()
    historic.save()
    bookmarks.save()

    self.quit()


# Run app

def main(version):
  app = Application()
  return app.run(sys.argv)




# header.py
#
# This file is part of Wike, a Wikipedia Reader for the GNOME Desktop.
# Copyright 2021 Hugo Olabera <hugolabe@gmail.com>.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import gi
gi.require_version('Gtk', '3.0')
gi.require_version('Handy', '1')
from gi.repository import Gtk, Handy

from wike.data import settings, historic
from wike.bookmarks import BookmarksPopover
from wike.langlinks import LanglinksPopover
from wike.menu import MenuPopover
from wike.search import SearchEntry
from wike.toc import TocPopover


# Headerbar for main window
# Contains widgets for manage searchs, navigation and popovers

@Gtk.Template(resource_path='/com/github/hugolabe/Wike/ui/header.ui')
class HeaderBar(Handy.HeaderBar):

  __gtype_name__ = 'HeaderBar'

  tab_button = Gtk.Template.Child()
  search_button = Gtk.Template.Child()
  menu_button = Gtk.Template.Child()
  bookmarks_button = Gtk.Template.Child()
  langlinks_button = Gtk.Template.Child()
  toc_button = Gtk.Template.Child()

  # Set main menu and connect signals and actions

  def __init__(self, window):
    super().__init__()

    self._window = window

    self.search_entry = SearchEntry(self.search_button, self._window)
    self.menu_popover = MenuPopover(self._window)
    self.bookmarks_popover = BookmarksPopover(self._window)
    self.langlinks_popover = LanglinksPopover(self._window)
    self.toc_popover = TocPopover(self._window)

    self.menu_button.set_popover(self.menu_popover)
    self.bookmarks_button.set_popover(self.bookmarks_popover)
    self.langlinks_button.set_popover(self.langlinks_popover)
    self.toc_button.set_popover(self.toc_popover)

    self.search_button.connect('clicked', self._search_button_cb)
    self.menu_button.connect('toggled', self._menu_button_cb)

  # Populate toc popover

  def set_toc(self, sections):
    if sections == None:
      self.toc_popover.populate(None)
    else:
      if len(sections) > 0:
        self.toc_popover.populate(sections)
        self.toc_button.set_sensitive(True)
      else:
        self.toc_popover.populate(None)

  # Populate langlinks popover

  def set_langlinks(self, langlinks):
    if langlinks == None:
      self.langlinks_popover.populate(None)
    else:
      if len(langlinks) > 0:
        if self.langlinks_popover.populate(langlinks):
          self.langlinks_button.set_sensitive(True)
      else:
        self.langlinks_popover.populate(None)

  # Refresh_languages for settings and langlinks popovers

  def refresh_langs(self):
    self.search_entry.settings_popover.populate_list()
    if self.search_entry.settings_popover.is_visible():
      self.search_entry.settings_popover.show_all()

    if self.langlinks_popover.refresh():
      if not self.langlinks_button.get_sensitive():
        self.langlinks_button.set_sensitive(True)
      if self.langlinks_popover.is_visible():
        self.langlinks_popover.show_all()
    else:
      if self.langlinks_popover.is_visible():
        self.langlinks_popover.hide()
      self.langlinks_button.set_sensitive(False)

  # Show search entry on button toggled

  def _search_button_cb(self, search_button):
    if search_button.get_active():
      self.set_custom_title(self.search_entry)
      self.search_entry.show()
      self.search_entry.grab_focus()
    else:
      self.search_entry.hide()
      self.set_custom_title(None)

  # Enable or disable menu items on button toggled

  def _menu_button_cb(self, menu_button):
    if menu_button.get_active():
      show_historic_action = self._window.lookup_action('show_historic')
      open_browser_action = self._window.lookup_action('open_browser')
      copy_url_action = self._window.lookup_action('copy_url')

      if self._window.page.wikiview.is_local():
        open_browser_action.set_enabled(False)
        copy_url_action.set_enabled(False)
      else:
        open_browser_action.set_enabled(True)
        copy_url_action.set_enabled(True)

      if settings.get_boolean('keep-historic') and len(historic.items) > 0:
        show_historic_action.set_enabled(True)
      else:
        show_historic_action.set_enabled(False)




# page.py
#
# This file is part of Wike, a Wikipedia Reader for the GNOME Desktop.
# Copyright 2021 Hugo Olabera <hugolabe@gmail.com>.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import gi
gi.require_version('Gtk', '3.0')
gi.require_version('WebKit2', '4.0')
from gi.repository import Gdk, Gtk, WebKit2

from wike.data import settings, historic
from wike.view import WikiView


# Page box for each tab
# Contains a webview and a search bar

@Gtk.Template(resource_path='/com/github/hugolabe/Wike/ui/page.ui')
class PageBox(Gtk.Box):

  __gtype_name__ = 'PageBox'

  search_bar = Gtk.Template.Child()
  textsearch_entry = Gtk.Template.Child()
  textsearch_prev_button = Gtk.Template.Child()
  textsearch_next_button = Gtk.Template.Child()
  textsearch_matches_label = Gtk.Template.Child()

  # Add wikiview, initialize find controller and connect signals

  def __init__(self, window):
    super().__init__()

    self._window = window

    self.wikiview = WikiView()
    self.pack_end(self.wikiview, True, True, 0)
    self.wikiview.show()

    find_controller = self.wikiview.get_find_controller()
    nav_list = self.wikiview.get_back_forward_list()
    self.search_bar.connect_entry(self.textsearch_entry)

    self.wikiview.connect('load-changed', self._wikiview_load_changed_cb)
    self.wikiview.connect('load-failed', self._wikiview_load_failed_cb)
    self.wikiview.connect('new-page', self._wikiview_new_page_cb)
    self.wikiview.connect('add_bookmark', self._wikiview_add_bookmark_cb)
    self.textsearch_entry.connect('changed', self._textsearch_entry_changed_cb, find_controller)
    self.textsearch_entry.connect('activate', self._textsearch_entry_activate_cb, find_controller)
    self.textsearch_prev_button.connect('clicked', self._textsearch_prev_button_clicked_cb, find_controller)
    self.textsearch_next_button.connect('clicked', self._textsearch_next_button_clicked_cb, find_controller)
    find_controller.connect('found-text', self._find_controller_found_cb)
    find_controller.connect('failed-to-find-text', self._find_controller_not_found_cb)
    find_controller.connect('counted-matches', self._find_controller_matches_cb)
    nav_list.connect('changed', self._nav_list_changed_cb)

  # Manage wikiview load page events

  def _wikiview_load_changed_cb(self, wikiview, event):
    tabpage = self._window.tabview.get_page(self)

    if event == WebKit2.LoadEvent.STARTED:
      if wikiview.get_uri().endswith('.wikipedia.org/'):
        self._is_main = True
      else:
        self._is_main = False
      if self.search_bar.get_search_mode(): self.search_bar.set_search_mode(False)
      tabpage.set_title(_('Loading Article'))
      tabpage.set_loading(True)
      if tabpage.get_selected():
        self._window.headerbar.set_title(_('Loading Article'))
        self._window.headerbar.toc_button.set_sensitive(False)
        self._window.headerbar.langlinks_button.set_sensitive(False)
        if self._window.headerbar.search_button.get_active():
          self._window.headerbar.search_button.set_active(False)
    elif event == WebKit2.LoadEvent.COMMITTED:
      wikiview.set_props()
      tabpage.set_title(wikiview.title)
      if tabpage.get_selected():
        self._window.headerbar.set_title(wikiview.title)
        self._window.headerbar.set_toc(wikiview.sections)
        self._window.headerbar.set_langlinks(wikiview.langlinks)
    elif event == WebKit2.LoadEvent.FINISHED:
      tabpage.set_loading(False)
      if settings.get_boolean('keep-historic'):
        if not self._is_main and not wikiview.is_local():
          historic.add(wikiview.get_base_uri(), wikiview.title, wikiview.get_lang())

  # If wikiview load failed show error

  def _wikiview_load_failed_cb(self, wikiview, event, uri, error):
    if not wikiview.is_loading():
      wikiview.load_message('error', uri)
    return True

  # On webview event create new page

  def _wikiview_new_page_cb(self, wikiview, uri):
    tabpage = self._window.tabview.get_page(self)
    self._window.new_page(uri, tabpage, False)

  # On webview event add new bookmark

  def _wikiview_add_bookmark_cb(self, wikiview, uri, title, lang):
    if self._window.headerbar.bookmarks_popover.add_bookmark(uri, title, lang):
      message = _('New bookmark: ') + title
      self._window.show_notification(message)

  # Search text in article when entry changes

  def _textsearch_entry_changed_cb(self, textsearch_entry, find_controller):
    text_length = textsearch_entry.get_text_length()
    if text_length > 2:
      text = textsearch_entry.get_text()
      find_controller.count_matches(text, WebKit2.FindOptions.WRAP_AROUND | WebKit2.FindOptions.CASE_INSENSITIVE, 9999)
      find_controller.search(text, WebKit2.FindOptions.WRAP_AROUND | WebKit2.FindOptions.CASE_INSENSITIVE, 9999)
    else:
      self.textsearch_matches_label.set_label('')
      self.textsearch_prev_button.set_sensitive(False)
      self.textsearch_next_button.set_sensitive(False)
      find_controller.search_finish()

  # On entry activated search next match

  def _textsearch_entry_activate_cb(self, textsearch_entry, find_controller):
    text_length = textsearch_entry.get_text_length()
    if text_length > 2:
      find_controller.search_next()

  # On text search prev button clicked search previous match

  def _textsearch_prev_button_clicked_cb(self, textsearch_prev_button, find_controller):
    find_controller.search_previous()

  # On text search next button clicked search next match

  def _textsearch_next_button_clicked_cb(self, textsearch_next_button, find_controller):
    find_controller.search_next()

  # Found text in article

  def _find_controller_found_cb(self, find_controller, match_count):
    self.textsearch_prev_button.set_sensitive(True)
    self.textsearch_next_button.set_sensitive(True)

  # Not found text in article

  def _find_controller_not_found_cb(self, find_controller):
    self.textsearch_matches_label.set_markup('<span foreground="red">' + str(0) + '</span>')
    self.textsearch_prev_button.set_sensitive(False)
    self.textsearch_next_button.set_sensitive(False)
    find_controller.search_finish()
    Gdk.beep()

  # Show text search matches found

  def _find_controller_matches_cb(self, find_controller, match_count):
    self.textsearch_matches_label.set_label(str(match_count))

  # Refresh navbox state on navlist changed

  def _nav_list_changed_cb(self, nav_list, item_added, item_removed):
    tabpage = self._window.tabview.get_page(self)
    if tabpage.get_selected():
      self._window.refresh_nav_buttons(self.wikiview)




# langlinks.py
#
# This file is part of Wike, a Wikipedia Reader for the GNOME Desktop.
# Copyright 2021 Hugo Olabera <hugolabe@gmail.com>.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import gi
gi.require_version('Gtk', '3.0')
from gi.repository import GLib, Gtk

from wike.data import languages


# Popover class for langlinks
# Show links to other languages for current page

@Gtk.Template(resource_path='/com/github/hugolabe/Wike/ui/langlinks.ui')
class LanglinksPopover(Gtk.Popover):

  __gtype_name__ = 'LanglinksPopover'

  langlinks_list = Gtk.Template.Child()
  filter_entry = Gtk.Template.Child()

  # Set list filter function for entry content and connect signals

  def __init__(self, window):
    super().__init__()

    self._window = window

    self._langlinks = None
    self.langlinks_list.set_filter_func(self._filter_list, self.filter_entry)
    self.langlinks_list.set_sort_func(self._sort_list)
    self.langlinks_list.set_header_func(self._set_row_separator)

    self.langlinks_list.connect('row-activated', self._list_activated_cb)
    self.filter_entry.connect('search-changed', self._filter_entry_changed_cb)
    self.connect('show', self._popover_show_cb)

  # Filter list for entry content

  def _filter_list(self, row, filter_entry):
    text = filter_entry.get_text()
    if text == '':
      return True

    if row.lang_name.lower().startswith(text.lower()) or row.lang.lower().startswith(text.lower()):
      return True
    else:
      return False

  # Sort list alphabetically

  def _sort_list(self, row1, row2):
    if row1.lang > row2.lang:
      return 1
    elif row1.lang < row2.lang:
      return -1
    else:
      return 0

  # Set langlink row separator

  def _set_row_separator(self, row, before):
    if before and not row.get_header():
      separator = Gtk.Separator.new(Gtk.Orientation.HORIZONTAL)
      row.set_header(separator)

  # Populate langlinks list

  def populate(self, langlinks):
    self._langlinks = langlinks

    return self.refresh()

  # Refresh langlinks list

  def refresh(self):
    rows = self.langlinks_list.get_children()
    for row in rows:
      self.langlinks_list.remove(row)

    if self._langlinks != None:
      for langlink in self._langlinks:
        if langlink['lang'] in languages.items:
          row = LanglinkBoxRow(langlink['url'], langlink['lang'], langlink['autonym'].capitalize(), langlink['*'])
          self.langlinks_list.add(row)

    if len(self.langlinks_list.get_children()) > 0:
      return True
    else:
      return False

  # Load article in view on list activated

  def _list_activated_cb(self, langlinks_list, row):
    self.hide()
    self._window.page.wikiview.load_wiki(row.uri)

  # Refresh list on filter entry changed

  def _filter_entry_changed_cb(self, filter_entry):
    self.langlinks_list.invalidate_filter()

  # Unselect items and show popover

  def _popover_show_cb(self, popover):
    self.langlinks_list.show_all()
    self.langlinks_list.unselect_all()


# Class for row in langlinks list
# Create row from uri, language and page title

@Gtk.Template(resource_path='/com/github/hugolabe/Wike/ui/langlinks-row.ui')
class LanglinkBoxRow(Gtk.ListBoxRow):

  __gtype_name__ = 'LanglinkBoxRow'

  langlink_lang = Gtk.Template.Child()
  langlink_title = Gtk.Template.Child()
  langlink_id = Gtk.Template.Child()

  # Set values

  def __init__(self, uri, lang, lang_name, title):
    super().__init__()

    self.uri = uri
    self.lang = lang
    self.lang_name = lang_name

    self.langlink_lang.set_label(lang_name)
    self.langlink_title.set_markup('<small>' + GLib.markup_escape_text(title, -1) + '</small>')
    self.langlink_id.set_label(lang)




# data.py
#
# This file is part of Wike, a Wikipedia Reader for the GNOME Desktop.
# Copyright 2021 Hugo Olabera <hugolabe@gmail.com>.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import datetime
import json
import os

import gi
from gi.repository import Gio, GLib


# Create app settings object and set data path variable

settings = Gio.Settings.new('com.github.hugolabe.Wike')
_data_path = GLib.get_user_data_dir()


# Class for Wikipedia languages
# wikilangs dict: all languages, items dict: filtered languages

class Languages:

  items = {}
  wikilangs = {}
  global _data_path
  _file_path = _data_path + '/languages.json'

  # Load all languages from resource and filtered languages from file

  def __init__(self):
    gfile = Gio.File.new_for_uri('resource:///com/github/hugolabe/Wike/languages/languages.min.json')
    try:
      gfile_contents = gfile.load_contents(None)
    except:
      print('Can\'t load languages file from resources')
      return
    else:
      content = gfile_contents[1].decode('utf-8')
      wikipedia_langs = json.loads(content)
      for lang_key in wikipedia_langs.keys():
        self.wikilangs[lang_key] = wikipedia_langs[lang_key]

    if os.path.exists(self._file_path):
      with open(self._file_path, 'r') as file:
        json_langs = json.load(file)
      for lang_key in json_langs.keys():
        self.items[lang_key] = json_langs[lang_key]
    else:
      self.items['en'] = 'English'

  # Clear languages list (filtered)

  def clear(self):
    self.items.clear()

  # Save json languages file

  def save(self):
    with open(self._file_path, 'w') as file:
      json.dump(self.items, file)


# Class for historic
# Maintains a list of viewed articles

class Historic:

  items = {}
  global _data_path
  _file_path = _data_path + '/historic.json'

  # Load json historic file

  def __init__(self):
    global settings

    if settings.get_boolean('keep-historic') and os.path.exists(self._file_path):
      with open(self._file_path, 'r') as file:
        json_historic = json.load(file)
        i = 0
        for item_key in sorted(json_historic.keys(), reverse=True):
          self.items[item_key] = json_historic[item_key]
          i += 1
          if i > 50: break

  # Add article to historic

  def add(self, uri, title, lang):
    now = datetime.datetime.today()
    date = now.strftime('%Y-%m-%d')
    time = now.strftime('%H:%M:%S')
    if date in self.items:
      self.items[date][uri] = [time, title, lang]
    else:
      self.items[date] = {uri: [time, title, lang]}

  # Clear historic

  def clear(self):
    self.items.clear()

  # Save json historic file

  def save(self):
    with open(self._file_path, 'w') as file:
      json.dump(self.items, file)


# Class for bookmarks
# List of articles marked for read later

class Bookmarks:

  items = {}
  global _data_path
  _file_path = _data_path + '/bookmarks.json'

  # Load json bookmarks file

  def __init__(self):
    if os.path.exists(self._file_path):
      with open(self._file_path, 'r') as file:
        json_bookmarks = json.load(file)
      for item_key in json_bookmarks.keys():
        self.items[item_key] = json_bookmarks[item_key]

  # Add article to bookmarks

  def add(self, uri, title, lang):
    if not uri in self.items:
      self.items[uri] = [title, lang]
      return True
    else:
      return False

  # Remove article from bookmarks

  def remove(self, uri):
    del self.items[uri]

  # Clear list of bookmarks

  def clear(self):
    self.items.clear()

  # Save json bookmarks file

  def save(self):
    with open(self._file_path, 'w') as file:
      json.dump(self.items, file)


# Create global objects for languages, historic and bookmarks

languages = Languages()
historic = Historic()
bookmarks = Bookmarks()




# window.py
#
# This file is part of Wike, a Wikipedia Reader for the GNOME Desktop.
# Copyright 2021 Hugo Olabera <hugolabe@gmail.com>.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import gi
gi.require_version('Gtk', '3.0')
gi.require_version('Handy', '1')
from gi.repository import Gio, GLib, Gdk, Gtk, Handy

from wike.data import settings
from wike.header import HeaderBar
from wike.page import PageBox


# Application main window
# Contains a page and a headerbar

@Gtk.Template(resource_path='/com/github/hugolabe/Wike/ui/window.ui')
class Window(Handy.ApplicationWindow):

  __gtype_name__ = 'Window'

  window_box = Gtk.Template.Child()
  tabbar = Gtk.Template.Child()
  tabview = Gtk.Template.Child()
  notification = Gtk.Template.Child()
  notification_label = Gtk.Template.Child()
  notification_close_button = Gtk.Template.Child()

  # Initialize window, set actions and connect signals

  def __init__(self, app, launch_uri):
    super().__init__(title='Wike', application=app)

    # Set default window icon name for XFCE, LXQt, MATE
    self.set_default_icon_name('com.github.hugolabe.Wike')

    self.set_default_size(settings.get_int('window-width'), settings.get_int('window-height'))
    if settings.get_boolean('window-max'):
      self.maximize()

    self.page = PageBox(self)
    tabpage = self.tabview.append(self.page)

    self.headerbar = HeaderBar(self)
    self.window_box.pack_start(self.headerbar, False, True, 0)

    actions = [ ('prev_page', self._prev_page_cb, ('<Alt>Left',)),
                ('next_page', self._next_page_cb, ('<Alt>Right',)),
                ('new_tab', self._new_tab_cb, ('<Ctrl>N',)),
                ('close_tab', self._close_tab_cb, ('<Ctrl>W',)),
                ('next_tab', self._next_tab_cb, ('<Ctrl>Tab',)),
                ('prev_tab', self._prev_tab_cb, ('<Shift><Ctrl>Tab',)),
                ('show_search', self._show_search_cb, ('F2', '<Ctrl>K',)),
                ('show_menu', self._show_menu_cb, ('F10',)),
                ('show_bookmarks', self._show_bookmarks_cb, ('<Ctrl>B',)),
                ('add_bookmark', self._add_bookmark_cb, ('<Ctrl>D',)),
                ('show_langlinks', self._show_langlinks_cb, ('<Ctrl>L',)),
                ('show_toc', self._show_toc_cb, ('<Ctrl>T',)),
                ('main_page', self._main_page_cb, ('<Alt>Home',)),
                ('random_article', self._random_article_cb, None),
                ('show_historic', self._show_historic_cb, ('<Ctrl>H',)),
                ('reload_page', self._reload_page_cb, ('F5', '<Ctrl>R',)),
                ('search_text', self._search_text_cb, ('<Ctrl>F',)),
                ('open_browser', self._open_browser_cb, None),
                ('copy_url', self._copy_url_cb, ('<Ctrl>U',)) ]

    for action, callback, accel in actions:
      simple_action = Gio.SimpleAction.new(action, None)
      simple_action.connect('activate', callback)
      self.add_action(simple_action)
      if accel:
        app.set_accels_for_action('win.' + action, accel)

    self.handler_selpage = self.tabview.connect('notify::selected-page', self._tabview_selected_page_cb)
    self.tabview.connect('close-page', self._tabview_close_page_cb)
    self.notification_close_button.connect('clicked', self._hide_notification_cb)

    if launch_uri == 'notfound':
      self.page.wikiview.load_message(launch_uri, None)
    else:
      if launch_uri != '':
        self.page.wikiview.load_wiki(launch_uri)
      else:
        if settings.get_int('on-start-load') == 0:
          self.page.wikiview.load_main()
        elif settings.get_int('on-start-load') == 1:
          self.page.wikiview.load_random()
        else:
          if settings.get_string('last-uri'):
            self.page.wikiview.load_wiki(settings.get_string('last-uri'))
          else:
            self.page.wikiview.load_main()

  # Create new tab with a page

  def new_page(self, uri, parent, select):
    page = PageBox(self)
    tabpage = self.tabview.add_page(page, parent)
    if uri == 'blank' or uri == 'notfound':
      page.wikiview.load_message(uri, None)
    else:
      page.wikiview.load_wiki(uri)

    if select:
      self.tabview.set_selected_page(tabpage)

  # New empty tab

  def _new_tab_cb(self, action, parameter):
    self.new_page('blank', None, True)

  # Close current tab

  def _close_tab_cb(self, action, parameter):
    if self.tabview.get_n_pages() > 1:
      tabpage = self.tabview.get_selected_page()
      self.tabview.close_page(tabpage)

  # Go to next tab

  def _next_tab_cb(self, action, parameter):
    num_pages = self.tabview.get_n_pages()
    if num_pages == 1:
      return

    tabpage = self.tabview.get_selected_page()
    if self.tabview.get_page_position(tabpage) < num_pages - 1:
      self.tabview.select_next_page()
    else:
      tabpage = self.tabview.get_nth_page(0)
      self.tabview.set_selected_page(tabpage)

  # Go to previous tab

  def _prev_tab_cb(self, action, parameter):
    num_pages = self.tabview.get_n_pages()
    if num_pages == 1:
      return

    tabpage = self.tabview.get_selected_page()
    if self.tabview.get_page_position(tabpage) > 0:
      self.tabview.select_previous_page()
    else:
      tabpage = self.tabview.get_nth_page(num_pages - 1)
      self.tabview.set_selected_page(tabpage)

  # On tab selected event refresh headerbar controls

  def _tabview_selected_page_cb(self, tabview, value):
    tabpage = tabview.get_selected_page()
    if not tabpage:
      return

    self.page = tabpage.get_child()
    self.refresh_nav_buttons(self.page.wikiview)
    self.headerbar.set_title(self.page.wikiview.title)
    self.headerbar.toc_button.set_sensitive(False)
    self.headerbar.langlinks_button.set_sensitive(False)
    self.headerbar.set_toc(self.page.wikiview.sections)
    self.headerbar.set_langlinks(self.page.wikiview.langlinks)

  # On tab closed event destroy wikiview and confirm

  def _tabview_close_page_cb(self, tabview, tabpage):
    page = tabpage.get_child()
    page.wikiview.destroy()
    tabview.close_page_finish(tabpage, True)
    return True

  # Go to previous page

  def _prev_page_cb(self, action, parameter):
    if self.page.wikiview.can_go_back():
      self.page.wikiview.go_back()

  # Go to next page

  def _next_page_cb(self, action, parameter):
    if self.page.wikiview.can_go_forward():
      self.page.wikiview.go_forward()

  # Show search entry

  def _show_search_cb(self, action, parameter):
    self.headerbar.search_button.set_active(not self.headerbar.search_button.get_active())

  # Show main menu

  def _show_menu_cb(self, action, parameter):
    self.headerbar.menu_button.set_active(not self.headerbar.menu_button.get_active())

  # Show bookmarks popover

  def _show_bookmarks_cb(self, action, parameter):
    if not self.headerbar.bookmarks_button.get_active():
      self.headerbar.bookmarks_button.set_active(True)

  # Add current page to bookmarks

  def _add_bookmark_cb(self, action, parameter):
    if not self.page.wikiview.is_local():
      uri = self.page.wikiview.get_base_uri()
      title = self.page.wikiview.title
      lang = self.page.wikiview.get_lang()
      if self.headerbar.bookmarks_popover.add_bookmark(uri, title, lang):
        if self.headerbar.bookmarks_popover.is_visible():
          self.headerbar.bookmarks_popover.bookmarks_list.show_all()
        else:
          message = _('New bookmark: ') + title
          self.show_notification(message)

  # Show langlinks popover

  def _show_langlinks_cb(self, action, parameter):
    if self.headerbar.langlinks_button.get_sensitive():
      if not self.headerbar.langlinks_button.get_active():
        self.headerbar.langlinks_button.set_active(True)

  # Show toc popover

  def _show_toc_cb(self, action, parameter):
    if self.headerbar.toc_button.get_sensitive():
      if not self.headerbar.toc_button.get_active():
        self.headerbar.toc_button.set_active(True)

  # Show Wikipedia main page

  def _main_page_cb(self, action, parameter):
    self.page.wikiview.load_main()

  # Show Wikipedia random article

  def _random_article_cb(self, action, parameter):
    self.page.wikiview.load_random()

  # Show historic

  def _show_historic_cb(self, action, parameter):
    self.page.wikiview.load_historic()

  # Reload article view

  def _reload_page_cb(self, action, parameter):
    self.page.wikiview.reload()

  # Show text search bar

  def _search_text_cb(self, action, parameter):
    if not self.page.search_bar.get_search_mode():
      self.page.search_bar.set_search_mode(True)
    else:
      self.page.textsearch_entry.grab_focus()

  # Open article in external browser

  def _open_browser_cb(self, action, parameter):
    uri = self.page.wikiview.get_base_uri()
    Gtk.show_uri(None, uri, Gdk.CURRENT_TIME)

  # Copy article URL to clipboard

  def _copy_url_cb(self, action, parameter):
    uri = self.page.wikiview.get_base_uri()
    clipboard = Gtk.Clipboard.get(Gdk.SELECTION_CLIPBOARD)
    clipboard.set_text(uri, -1)

    self.show_notification(_('Wikipedia URL copied to clipboard'))

  # Refresh navigation buttons state

  def refresh_nav_buttons(self, wikiview):
    prev_page_action = self.lookup_action('prev_page')
    next_page_action = self.lookup_action('next_page')

    if wikiview.can_go_back():
      prev_page_action.set_enabled(True)
    else:
      prev_page_action.set_enabled(False)
    if wikiview.can_go_forward():
      next_page_action.set_enabled(True)
    else:
      next_page_action.set_enabled(False)

  # Show notification with provided message

  def show_notification(self, message):
    self.notification_label.set_label(message)
    self.notification.set_reveal_child(True)
    GLib.timeout_add_seconds(3, self._hide_notification_cb, None)

  # Hide notification on button clicked or time expired

  def _hide_notification_cb(self, close_button):
    if self.notification.get_reveal_child():
      self.notification.set_reveal_child(False)
    return False




# bookmarks.py
#
# This file is part of Wike, a Wikipedia Reader for the GNOME Desktop.
# Copyright 2021 Hugo Olabera <hugolabe@gmail.com>.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk

from wike.data import languages, bookmarks


# Popover class for bookmarks
# View and manage list of bookmarks

@Gtk.Template(resource_path='/com/github/hugolabe/Wike/ui/bookmarks.ui')
class BookmarksPopover(Gtk.Popover):

  __gtype_name__ = 'BookmarksPopover'

  bookmarks_list = Gtk.Template.Child()
  add_button = Gtk.Template.Child()
  clear_button = Gtk.Template.Child()
  placeholder_label = Gtk.Template.Child()

  # Initialize popover and connect signals

  def __init__(self, window):
    super().__init__()

    self._window = window

    placeholder_text = _("<b>No Bookmarks</b>\n\n<small>Use the <i>Add Bookmark</i> button\nor press Ctrl+D to add\nthe current article to the list</small>")
    self.placeholder_label.set_markup(placeholder_text)

    self.bookmarks_list.set_sort_func(self._sort_list)
    self.bookmarks_list.set_header_func(self._set_row_separator)
    self._populate()

    self.bookmarks_list.connect('row-activated', self._list_activated_cb)
    self.bookmarks_list.connect('add', self._list_changed_cb)
    self.bookmarks_list.connect('remove', self._list_changed_cb)
    self.add_button.connect('clicked', self._add_button_clicked_cb)
    self.clear_button.connect('clicked', self._clear_button_clicked_cb)
    self.connect('show', self._popover_show_cb)

  # Sort list alphabetically

  def _sort_list(self, row1, row2):
    if row1.title > row2.title:
      return 1
    elif row1.title < row2.title:
      return -1
    else:
      if row1.lang > row2.lang:
        return 1
      elif row1.lang < row2.lang:
        return -1
      else:
        return 0

  # Set bookmark row separator

  def _set_row_separator(self, row, before):
    if before and not row.get_header():
      separator = Gtk.Separator.new(Gtk.Orientation.HORIZONTAL)
      row.set_header(separator)

  # Populate bookmarks list

  def _populate(self):
    for bookmark in bookmarks.items:
      title = bookmarks.items[bookmark][0]
      lang = bookmarks.items[bookmark][1]
      if lang in languages.wikilangs:
        lang_name = languages.wikilangs[lang].capitalize()
      else:
        lang_name = lang
      row = BookmarkBoxRow(bookmark, title, lang, lang_name)
      self.bookmarks_list.add(row)

  # Refresh sensitivity of popover buttons

  def _refresh_buttons(self):
    if len(self.bookmarks_list.get_children()) == 0:
      self.clear_button.set_sensitive(False)
    else:
      self.clear_button.set_sensitive(True)
    if self._window.page.wikiview.get_base_uri() in bookmarks.items or self._window.page.wikiview.is_local():
      self.add_button.set_sensitive(False)
    else:
      self.add_button.set_sensitive(True)

  # Add bookmark with uri, title and lang

  def add_bookmark(self, uri, title, lang):
    if bookmarks.add(uri, title, lang):
      if lang in languages.wikilangs:
        lang_name = languages.wikilangs[lang].capitalize()
      else:
        lang_name = lang
      row = BookmarkBoxRow(uri, title, lang, lang_name)
      self.bookmarks_list.add(row)
      return True
    else:
      return False

  # Add item on button clicked

  def _add_button_clicked_cb(self, add_button):
    uri = self._window.page.wikiview.get_base_uri()
    title = self._window.page.wikiview.title
    lang = self._window.page.wikiview.get_lang()
    if self.add_bookmark(uri, title, lang):
      self.bookmarks_list.show_all()

  # Clear list on button clicked

  def _clear_button_clicked_cb(self, clear_button):
    clear_dialog = Gtk.MessageDialog(self.get_toplevel(),
                                     Gtk.DialogFlags.MODAL,
                                     Gtk.MessageType.WARNING,
                                     Gtk.ButtonsType.CANCEL,
                                     _('Clear Bookmarks List?'))

    clear_dialog.set_property('secondary-text', _('All bookmarks in the list will be removed permanently.'))
    clear_dialog.add_buttons(_('Clear List'), Gtk.ResponseType.YES)

    response = clear_dialog.run()
    clear_dialog.destroy()
    if response == Gtk.ResponseType.CANCEL:
      return

    bookmarks.clear()
    rows = self.bookmarks_list.get_children()
    for row in rows:
      self.bookmarks_list.remove(row)

  # Load article in view on list activated

  def _list_activated_cb(self, bookmarks_list, row):
    self.hide()
    self._window.page.wikiview.load_wiki(row.uri)

  # Refresh buttons state on list changed

  def _list_changed_cb(self, bookmarks_list, row):
    self._refresh_buttons()

  # Populate list and show popover

  def _popover_show_cb(self, popover):
    self.bookmarks_list.show_all()
    self.bookmarks_list.unselect_all()
    self._refresh_buttons()


# Class for row in bookmarks list
# Create row from URL, title and article language

@Gtk.Template(resource_path='/com/github/hugolabe/Wike/ui/bookmarks-row.ui')
class BookmarkBoxRow(Gtk.ListBoxRow):

  __gtype_name__ = 'BookmarkBoxRow'

  article_title = Gtk.Template.Child()
  article_lang = Gtk.Template.Child()
  remove_button = Gtk.Template.Child()

  # Set values and connect signals

  def __init__(self, uri, title, lang, lang_name):
    super().__init__()

    self.uri = uri
    self.title = title
    self.lang = lang

    self.article_title.set_label(title)
    self.article_lang.set_markup('<small>' + lang_name + '</small>')

    self.remove_button.connect('clicked', self._remove_button_clicked_cb)

  # Remove bookmark on buttom clicked

  def _remove_button_clicked_cb(self, remove_button):
    bookmarks.remove(self.uri)
    bookmarks_list = self.get_parent()
    bookmarks_list.remove(self)




# view.py
#
# This file is part of Wike, a Wikipedia Reader for the GNOME Desktop.
# Copyright 2021 Hugo Olabera <hugolabe@gmail.com>.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import urllib.parse

import gi
gi.require_version('Gtk', '3.0')
gi.require_version('WebKit2', '4.0')
from gi.repository import Gio, GLib, GObject, Gdk, Gtk, Handy, WebKit2

from wike import wikipedia
from wike.data import settings, languages, historic, bookmarks


# Initialize webcontext with cookie manager

web_context = WebKit2.WebContext.get_default()
web_context.set_cache_model(WebKit2.CacheModel.WEB_BROWSER)
cookies_file_path = GLib.get_user_data_dir() + '/cookies'
cookie_manager = WebKit2.WebContext.get_cookie_manager(web_context)
cookie_manager.set_accept_policy(WebKit2.CookieAcceptPolicy.ALWAYS)
cookie_manager.set_persistent_storage(cookies_file_path, WebKit2.CookiePersistentStorage.TEXT)


# Settings class for Wikipedia views
# Create settings and user content with custom css for wikiviews

class ViewSettings:

  web_settings = WebKit2.Settings()
  user_content = WebKit2.UserContentManager()
  style_manager = Handy.StyleManager.get_default()

  # Load custom css and connect view settings signals

  def __init__(self):
    self.web_settings.set_default_font_size(settings.get_int('font-size'))

    gfile = Gio.File.new_for_uri('resource:///com/github/hugolabe/Wike/styles/view.min.css')
    try:
      gfile_contents = gfile.load_contents(None)
    except:
      print('Can\'t load view css file from resources')
      self._css_view = ''
    else:
      self._css_view = gfile_contents[1].decode('utf-8')

    gfile = Gio.File.new_for_uri('resource:///com/github/hugolabe/Wike/styles/dark.min.css')
    try:
      gfile_contents = gfile.load_contents(None)
    except:
      print('Can\'t load dark css file from resources')
      self._css_dark = ''
    else:
      self._css_dark = gfile_contents[1].decode('utf-8')

    gfile = Gio.File.new_for_uri('resource:///com/github/hugolabe/Wike/styles/sepia.min.css')
    try:
      gfile_contents = gfile.load_contents(None)
    except:
      print('Can\'t load sepia css file from resources')
      self._css_sepia = ''
    else:
      self._css_sepia = gfile_contents[1].decode('utf-8')

    self.set_style()

    settings.connect('changed::font-size', self._settings_font_size_changed_cb)
    settings.connect('changed::custom-font', self._settings_custom_font_changed_cb)
    settings.connect('changed::font-family', self._settings_custom_font_changed_cb)
    settings.connect('changed::preview-popups', self._settings_preview_popups_changed_cb)
    settings.connect('changed::theme', self._settings_theme_changed_cb)
    self.style_manager.connect('notify::dark', self._system_theme_changed_cb)

  # Settings font-size changed event

  def _settings_font_size_changed_cb(self, settings, key):
    self.web_settings.set_default_font_size(settings.get_int('font-size'))

  # Settings custom font changed event

  def _settings_custom_font_changed_cb(self, settings, key):
    self.set_style()

  # Settings preview popups changed event

  def _settings_preview_popups_changed_cb(self, settings, key):
    self.set_style()

  # Settings theme changed event

  def _settings_theme_changed_cb(self, settings, key):
    self.set_style()

  # System theme changed event

  def _system_theme_changed_cb(self, style_manager, dark):
    self.set_style()

  # Inject stylesheets for customize article view

  def set_style(self):
    data_manager = web_context.get_website_data_manager()
    data_manager.clear(WebKit2.WebsiteDataTypes.MEMORY_CACHE , 0, None, None, None)
    self.user_content.remove_all_style_sheets()

    style_view = WebKit2.UserStyleSheet(self._css_view, WebKit2.UserContentInjectedFrames.ALL_FRAMES, WebKit2.UserStyleLevel.USER, None, None)
    self.user_content.add_style_sheet(style_view)

    theme = settings.get_int('theme')
    if theme == 1:
      style_dark = WebKit2.UserStyleSheet(self._css_dark, WebKit2.UserContentInjectedFrames.ALL_FRAMES, WebKit2.UserStyleLevel.USER, None, None)
      self.user_content.add_style_sheet(style_dark)
    elif theme == 2:
      style_sepia = WebKit2.UserStyleSheet(self._css_sepia, WebKit2.UserContentInjectedFrames.ALL_FRAMES, WebKit2.UserStyleLevel.USER, None, None)
      self.user_content.add_style_sheet(style_sepia)
    elif theme == 3:
      if self.style_manager.get_dark():
        style_dark = WebKit2.UserStyleSheet(self._css_dark, WebKit2.UserContentInjectedFrames.ALL_FRAMES, WebKit2.UserStyleLevel.USER, None, None)
        self.user_content.add_style_sheet(style_dark)

    if settings.get_boolean('custom-font'):
      css_font = 'body,h1,h2{font-family:"' + settings.get_string('font-family') + '"!important}'
      style_font = WebKit2.UserStyleSheet(css_font, WebKit2.UserContentInjectedFrames.ALL_FRAMES, WebKit2.UserStyleLevel.USER, None, None)
      self.user_content.add_style_sheet(style_font)

    if not settings.get_boolean('preview-popups'):
      css_previews = '.mwe-popups{display:none!important}'
      style_previews = WebKit2.UserStyleSheet(css_previews, WebKit2.UserContentInjectedFrames.ALL_FRAMES, WebKit2.UserStyleLevel.USER, None, None)
      self.user_content.add_style_sheet(style_previews)


# Create view_settings object

view_settings = ViewSettings()


# Webview class for Wikipedia views
# Show Wikipedia pages and manage decision policy

class WikiView(WebKit2.WebView):

  __gsignals__ = { 'new-page': (GObject.SIGNAL_RUN_FIRST, None, (str, )),
                   'add-bookmark': (GObject.SIGNAL_RUN_FIRST, None, (str, str, str,)) }

  title = 'Wike'
  sections = None
  langlinks = None

  # Initialize view with user content and web settings

  def __init__(self):
    super().__init__(settings=view_settings.web_settings, user_content_manager=view_settings.user_content)

    theme = settings.get_int('theme')
    if theme == 1:
      self.set_background_color(Gdk.RGBA(0.1, 0.1, 0.1, 1))
    elif theme == 2:
      self.set_background_color(Gdk.RGBA(0.976, 0.953, 0.914, 1))
    elif theme == 3:
      if view_settings.style_manager.get_dark():
        self.set_background_color(Gdk.RGBA(0.1, 0.1, 0.1, 1))
      else:
        self.set_background_color(Gdk.RGBA(1, 1, 1, 1))
    else:
      self.set_background_color(Gdk.RGBA(1, 1, 1, 1))

  # Load Wikipedia article by URI

  def load_wiki(self, uri):
    self.stop_loading()
    if uri.find('.m.') == -1:
      uri = uri.replace('.wikipedia.org', '.m.wikipedia.org')
    self.load_uri(uri)

  # Go to section in page

  def load_section(self, anchor):
    section = urllib.parse.quote(anchor)
    uri = self.get_uri()
    uri_elements = urllib.parse.urlparse(uri)
    if uri_elements[5] != section:
      section_uri_elements = (uri_elements[0], uri_elements[1], uri_elements[2], uri_elements[3], uri_elements[4], section)
      section_uri = urllib.parse.urlunparse(section_uri_elements)
      self.load_uri(section_uri)

  # Load Wikipedia main page

  def load_main(self):
    uri = 'https://' + settings.get_string('search-language') + '.m.wikipedia.org'
    self.load_wiki(uri)

  # Load Wikipedia random article

  def load_random(self):
    try:
      uri = wikipedia.get_random(settings.get_string('search-language'))
      self.load_wiki(uri)
    except:
      self.load_message('notfound', None)

  # Create and load html for message page

  def load_message(self, message_type, fail_uri):
    if message_type == 'blank':
      title = 'Wike'
      subtitle = _('Search and read Wikipedia articles')
      message = _('Click the search button or press F2 to search articles in Wikipedia. You can also use the main menu to open Wikipedia main page, the list of recent articles or get a random article.')
    elif message_type == 'notfound':
      title = ':-('
      subtitle = _('Article not found')
      message = _('Can\'t find any results for requested query')
    else:
      title = ':-('
      subtitle = _('Can\'t access Wikipedia')
      message = _('Check your Internet connection and try again')

    if fail_uri:
      button = '<a href="' + fail_uri + '" class="btn">' + _('Try Again') + '</a>\n'
    else:
      button = ''

    html = '<!DOCTYPE html>\n<html>\n<head>\n<meta http-equiv="Content-Type" content="text/html; charset=UTF-8" />\n<title>Wike</title>\n</head>\n\
<body class="wike-body">\n<div id="wike-message">\n<h1>' + title + '</h1>\n<h3>' + subtitle + '</h3>\n<p>' + message + '</p>\n' + button + '</div>\n</body>\n</html>\n'

    uri = 'about:' + message_type
    self.load_html(html, uri)

  # Create and load html for historic page

  def load_historic(self):
    html_top = '<!DOCTYPE html>\n<html>\n<head>\n<meta http-equiv="Content-Type" content="text/html; charset=UTF-8" />\n<title>Wike</title>\n</head>\n\
<body class="wike-body">\n<div id="wike-historic">\n'

    html_bottom = '</div>\n</body>\n</html>\n'

    html_body = ''
    for date in sorted(historic.items, reverse=True):
      html_body = html_body + '<h3>' + date + '</h3>\n<table>\n'
      historic_item = historic.items[date]
      for uri in sorted(historic_item, key=historic_item.get, reverse=True):
        time = historic_item[uri][0]
        title = historic_item[uri][1]
        lang = historic_item[uri][2]
        if lang in languages.wikilangs:
          lang = languages.wikilangs[lang].capitalize()
        html_body = html_body + '<tr>\n<td class="wike-hist-time">' + time.rsplit(':', 1)[0] + '</td>\n<td class="wike-hist-link">' + '<a href="' + uri + '">' + title + '</a></td>\n<td class="wike-hist-lang">' + lang + '</td>\n</tr>\n'
      html_body = html_body + '</table>\n'

    html = html_top + html_body + html_bottom

    uri = 'about:historic'
    self.load_alternate_html(html, uri, uri)

  # Get base uri for current article

  def get_base_uri(self):
    uri = self.get_uri().replace('.m.', '.')
    uri_elements = urllib.parse.urlparse(uri)
    if uri_elements[5]:
      base_uri_elements = (uri_elements[0], uri_elements[1], uri_elements[2], '', '', '')
      base_uri = urllib.parse.urlunparse(base_uri_elements)
      return base_uri
    else:
      return uri

  # Get language for current article

  def get_lang(self):
    uri = self.get_uri()
    uri_elements = urllib.parse.urlparse(uri)
    uri_scheme = uri_elements[0]
    uri_netloc = uri_elements[1]

    if uri_scheme == 'about':
      return None
    else:
      lang = uri_netloc.split('.', 1)[0]
      return lang

  # Set TOC, langlinks and title for current article

  def set_props(self):
    uri = self.get_uri()
    uri_elements = urllib.parse.urlparse(uri)
    uri_scheme = uri_elements[0]
    uri_netloc = uri_elements[1]
    uri_path = uri_elements[2]

    props = None
    self.title = 'Wike'
    self.sections = None
    self.langlinks = None

    if uri_scheme == 'about':
      if uri_path == 'historic':
        self.title = _('Recent Articles')
      elif uri_path == 'notfound':
        self.title = _('Article not found')
      elif uri_path == 'error':
        self.title = _('Can\'t access Wikipedia')
      return
    else:
      page = uri_path.replace('/wiki/', '', 1)
      lang = uri_netloc.split('.', 1)[0]
      try:
        props = wikipedia.get_properties(page, lang)
      except:
        props = None

    if props:
      self.title = props['title']
      self.sections = props['sections']
      self.langlinks = props['langlinks']
    else:
      title_raw = page.replace('_', ' ')
      self.title = urllib.parse.unquote(title_raw)

  # Check if current page is local (message page or historic)

  def is_local(self):
    uri = self.get_uri()
    uri_elements = urllib.parse.urlparse(uri)
    uri_scheme = uri_elements[0]

    if uri_scheme == 'about':
      return True
    else:
      return False

  # Check if uri is a valid Wikipedia URL

  def _is_wiki_uri(self, uri):
    uri_elements = urllib.parse.urlparse(uri)
    uri_scheme = uri_elements[0]
    uri_netloc = uri_elements[1]
    uri_path = uri_elements[2]

    if uri_netloc.endswith('.wikipedia.org') and (uri_path.startswith('/wiki/') or uri_path == '/'):
      base_uri_elements = (uri_scheme, uri_netloc.replace('.m.', '.'), uri_path, '', '', '')
      base_uri = urllib.parse.urlunparse(base_uri_elements)
      return base_uri
    else:
      return ''

  # Webview decision policy event

  def do_decide_policy(self, decision, decision_type):
    if decision_type == WebKit2.PolicyDecisionType.NAVIGATION_ACTION or decision_type == WebKit2.PolicyDecisionType.NEW_WINDOW_ACTION:
      nav_action = decision.get_navigation_action()
      nav_type = nav_action.get_navigation_type()
      mouse_button = nav_action.get_mouse_button()
      uri = nav_action.get_request().get_uri()
      uri_elements = urllib.parse.urlparse(uri)
      uri_scheme = uri_elements[0]
      uri_netloc = uri_elements[1]
      uri_path = uri_elements[2]
      uri_fragment = uri_elements[5]
      if nav_type == WebKit2.NavigationType.LINK_CLICKED:
        if uri_netloc.endswith('.wikipedia.org') and (uri_path.startswith('/wiki/') or uri_path == '/'):
          base_uri_elements = (uri_elements[0], uri_elements[1].replace('.m.', '.'), uri_elements[2], '', '', '')
          base_uri = urllib.parse.urlunparse(base_uri_elements)
          if mouse_button == 2:
            decision.ignore()
            self.emit('new-page', base_uri)
          else:
            if base_uri != self.get_base_uri():
              decision.ignore()
              self.load_wiki(uri)
        else:
          decision.ignore()
          Gtk.show_uri(None, uri, Gdk.CURRENT_TIME)
      elif nav_type == WebKit2.NavigationType.RELOAD or nav_type == WebKit2.NavigationType.BACK_FORWARD:
        if uri_scheme == 'about':
          decision.ignore()
          if uri_path == 'historic':
            self.load_historic()
      elif nav_type == WebKit2.NavigationType.OTHER:
        if uri_netloc.startswith('upload.'):
          decision.ignore()
    return True

  # Manage webview context menu

  def do_context_menu(self, menu, event, hit_test):
    if hit_test.context_is_editable() or hit_test.context_is_media() or hit_test.context_is_scrollbar():
      return True

    action_tab = Gio.SimpleAction.new('new_tab', GLib.VariantType('s'))
    action_tab.connect('activate', self._new_tab_cb)
    action_browser = Gio.SimpleAction.new('open_browser', GLib.VariantType('s'))
    action_browser.connect('activate', self._open_browser_cb)
    action_bookmark = Gio.SimpleAction.new('add_bookmark', GLib.VariantType('s'))
    action_bookmark.connect('activate', self._add_bookmark_cb)
    action_clipboard = Gio.SimpleAction.new('copy_link', GLib.VariantType('s'))
    action_clipboard.connect('activate', self._copy_link_cb)

    if hit_test.get_context() == WebKit2.HitTestResultContext.DOCUMENT:
      base_uri = self.get_base_uri()
      uri = self._is_wiki_uri(base_uri)
      if uri == '' or uri in bookmarks.items:
        action_bookmark.set_enabled(False)

      item = WebKit2.ContextMenuItem.new_separator()
      menu.append(item)
      item = WebKit2.ContextMenuItem.new_from_gaction(action_bookmark, _('Add Bookmark'), GLib.Variant.new_string(uri))
      menu.append(item)

    if hit_test.context_is_link() and not hit_test.context_is_image():
      link = hit_test.get_link_uri()
      uri = self._is_wiki_uri(link)
      if uri == '':
        action_bookmark.set_enabled(False)
        action_tab.set_enabled(False)
      else:
        if uri in bookmarks.items:
          action_bookmark.set_enabled(False)

      if uri != '':
        link = uri
      item = WebKit2.ContextMenuItem.new_from_gaction(action_tab, _('Open Link in New Tab'), GLib.Variant.new_string(uri))
      menu.append(item)
      item = WebKit2.ContextMenuItem.new_from_gaction(action_browser, _('Open Link in Browser'), GLib.Variant.new_string(link))
      menu.append(item)
      item = WebKit2.ContextMenuItem.new_separator()
      menu.append(item)
      item = WebKit2.ContextMenuItem.new_from_gaction(action_bookmark, _('Add Link to Bookmarks'), GLib.Variant.new_string(uri))
      menu.append(item)
      item = WebKit2.ContextMenuItem.new_from_gaction(action_clipboard, _('Copy Link to Clipboard'), GLib.Variant.new_string(link))
      menu.append(item)

    allowed_actions = (WebKit2.ContextMenuAction.NO_ACTION,
                       WebKit2.ContextMenuAction.CUSTOM,
                       WebKit2.ContextMenuAction.GO_BACK,
                       WebKit2.ContextMenuAction.GO_FORWARD,
                       WebKit2.ContextMenuAction.RELOAD,
                       WebKit2.ContextMenuAction.COPY,
                       WebKit2.ContextMenuAction.COPY_IMAGE_TO_CLIPBOARD,
                       WebKit2.ContextMenuAction.COPY_IMAGE_URL_TO_CLIPBOARD)

    items = menu.get_items()
    for item in items:
      if not item.get_stock_action() in allowed_actions:
        menu.remove(item)

    return False

  # On context menu activated open link in new tab

  def _new_tab_cb(self, action, parameter):
    uri = parameter.get_string()
    self.emit('new-page', uri)

  # On context menu activated open link in browser

  def _open_browser_cb(self, action, parameter):
    uri = parameter.get_string()
    Gtk.show_uri(None, uri, Gdk.CURRENT_TIME)

  # On context menu activated add bookmark

  def _add_bookmark_cb(self, action, parameter):
    uri = parameter.get_string()
    uri_elements = urllib.parse.urlparse(uri)
    uri_netloc = uri_elements[1]
    uri_path = uri_elements[2]

    page = uri_path.replace('/wiki/', '', 1)
    title_raw = page.replace('_', ' ')
    title = urllib.parse.unquote(title_raw)
    lang = uri_netloc.split('.', 1)[0]

    self.emit('add-bookmark', uri, title, lang)

  # On context menu activated copy link to clipboard

  def _copy_link_cb(self, action, parameter):
    uri = parameter.get_string()
    clipboard = Gtk.Clipboard.get(Gdk.SELECTION_CLIPBOARD)
    clipboard.set_text(uri, -1)




# wikipedia.py
#
# This file is part of Wike, a Wikipedia Reader for the GNOME Desktop.
# Copyright 2021 Hugo Olabera <hugolabe@gmail.com>.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import requests


# create a session for wikipedia requests

_session = requests.Session()

# Get Wikipedia random page

def get_random(lang):
  api = 'https://' + lang + '.wikipedia.org/w/api.php'
  params = { 'action': 'query',
             'generator': 'random',
             'grnlimit': 1,
             'grnnamespace': 0,
             'prop': 'info',
             'inprop': 'url',
             'format': 'json' }

  result = _request(api, params)
  pages = result['query']['pages']
  for page, info in pages.items():
    uri = info['fullurl']

  return uri

# Search in Wikipedia with a limit of responses

def search(text, lang, limit):
  api = 'https://' + lang + '.wikipedia.org/w/api.php'
  params = { 'action': 'opensearch',
             'search': text,
             'limit': limit,
             'namespace': 0,
             'redirects': 'resolve',
             'format': 'json' }

  result = _request(api, params)

  if len(result[1]) == 0:
    return None
  else:
    return result[1], result[3]

# Get sections and langlinks for Wikipedia page

def get_properties(page, lang):
  api = 'https://' + lang + '.wikipedia.org/w/api.php'
  params = '?format=json&action=parse&prop=sections|langlinks&redirects&page=' + page

  result = _request(api, params)

  return result['parse']

# Perform query to Wikipedia API with given parameters

def _request(api, params):
  global _session
  headers = { 'User-Agent': 'Wike/1.8.2 (https://github.com/hugolabe)' }

  if type(params) is dict:
    response = _session.get(url=api, params=params, headers=headers, timeout=(4, 12))
  else:
    uri = api + params
    response = _session.get(url=uri, params=None, headers=headers, timeout=(4, 12))

  return response.json()




# toc.py
#
# This file is part of Wike, a Wikipedia Reader for the GNOME Desktop.
# Copyright 2021 Hugo Olabera <hugolabe@gmail.com>.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gio, GLib, Gtk


# Popover class for TOC
# Show table of contents for current page

class TocPopover(Gtk.Popover):

  # Initialize popover and set actions

  def __init__(self, window):
    super().__init__()
    self.set_size_request(300, -1)

    self._window = window

    actions = Gio.SimpleActionGroup()
    action = Gio.SimpleAction.new('section', GLib.VariantType('s'))
    action.connect('activate', self._section_activate_cb)
    actions.add_action(action)
    self.insert_action_group('toc_popover', actions)

  # Create toc menu and populate

  def populate(self, toc):
    self._toc = toc
    menu = Gio.Menu()
    self.bind_model(menu, None)
    if toc != None:
      self._fill_menu(menu, toc, 0)

  # Fill toc menu and submenus

  def _fill_menu(self, menu, toc, index):
    level = toc[index]['toclevel']

    if index > 0:
      action_string = 'toc_popover.section(\'' + str(index-1) + '\')'
      text = toc[index-1]['anchor'].replace('_', ' ')
      if len(text) > 45:
        text = text[0:44] + '...'
      button = Gio.MenuItem.new(text, action_string)
      menu.append_item(button)

    while index < len(toc) - 1:
      leap = level - toc[index+1]['toclevel']
      text = toc[index]['anchor'].replace('_', ' ')
      if len(text) > 45:
        text = text[0:44] + '...'

      if leap >= 0:
        action_string = 'toc_popover.section(\'' + str(index) + '\')'
        button = Gio.MenuItem.new(text, action_string)
        menu.append_item(button)
        index += 1
        if leap > 0:
          return index
      elif leap < 0:
        submenu = Gio.Menu()
        menu.append_submenu(text, submenu)
        index = self._fill_menu(submenu, toc, index+1)
        if toc[index]['toclevel'] < level:
          return index

    if toc[index]['toclevel'] == level:
      action_string = 'toc_popover.section(\'' + str(index) + '\')'
      text = toc[index]['anchor'].replace('_', ' ')
      if len(text) > 45:
        text = text[0:44] + '...'
      button = Gio.MenuItem.new(text, action_string)
      menu.append_item(button)
      return index

    return index

  # On menu activated load section

  def _section_activate_cb(self, action, parameter):
    index = int(parameter.unpack())
    anchor = self._toc[index]['anchor']
    self._window.page.wikiview.load_section(anchor)




# prefs.py
#
# This file is part of Wike, a Wikipedia Reader for the GNOME Desktop.
# Copyright 2021 Hugo Olabera <hugolabe@gmail.com>.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import gi
gi.require_version('Gtk', '3.0')
gi.require_version('Handy', '1')
gi.require_version('WebKit2', '4.0')
from gi.repository import Gio, Gtk, Handy, WebKit2

from wike.data import settings, languages, historic


# Preferences window
# Manage application preferences

@Gtk.Template(resource_path='/com/github/hugolabe/Wike/ui/prefs.ui')
class PrefsWindow(Handy.PreferencesWindow):

  __gtype_name__ = 'PrefsWindow'

  on_start_combo = Gtk.Template.Child()
  font_size_spin = Gtk.Template.Child()
  custom_font_switch = Gtk.Template.Child()
  custom_font_combo = Gtk.Template.Child()
  preview_popups_switch = Gtk.Template.Child()
  desktop_search_switch = Gtk.Template.Child()
  historic_switch = Gtk.Template.Child()
  clear_historic_button = Gtk.Template.Child()
  clear_data_button = Gtk.Template.Child()
  languages_list = Gtk.Template.Child()
  select_all_button = Gtk.Template.Child()
  select_none_button = Gtk.Template.Child()

  # Connect signals and bindings

  def __init__(self):
    super().__init__()

    self._languages_changed = False
    self._populate_on_start_combo()
    self._populate_custom_font_combo()
    self._populate_languages_list()

    settings.bind('on-start-load', self.on_start_combo, 'selected-index', Gio.SettingsBindFlags.DEFAULT)
    settings.bind('font-size', self.font_size_spin, 'value', Gio.SettingsBindFlags.DEFAULT)
    settings.bind('custom-font', self.custom_font_switch, 'active', Gio.SettingsBindFlags.DEFAULT)
    settings.bind('preview-popups', self.preview_popups_switch, 'active', Gio.SettingsBindFlags.DEFAULT)
    settings.bind('search-desktop', self.desktop_search_switch, 'active', Gio.SettingsBindFlags.DEFAULT)
    settings.bind('keep-historic', self.historic_switch, 'active', Gio.SettingsBindFlags.DEFAULT)

    self.custom_font_combo.connect('notify::selected-index', self._custom_font_combo_selected_cb)
    self.clear_historic_button.connect('clicked', self._clear_historic_button_cb)
    self.clear_data_button.connect('clicked', self._clear_data_button_cb)
    self.languages_list.connect('row-activated', self._languages_list_selected_cb)
    self.select_all_button.connect('clicked', self._select_all_button_cb)
    self.select_none_button.connect('clicked', self._select_none_button_cb)

    self.connect('delete-event', self._window_delete_cb)

  # Populate list of available options for start page

  def _populate_on_start_combo(self):
    model = Gio.ListStore.new(Handy.ValueObject)
    options = (_('Wikipedia Main Page'), _('Random Article'), _('Last Article'))
    for index, option in enumerate(options):
      model.insert(index, Handy.ValueObject.new(option))
    self.on_start_combo.bind_name_model(model, Handy.ValueObject.dup_string)

  # Populate font list for custom font combo

  def _populate_custom_font_combo(self):
    model = Gio.ListStore.new(Handy.ValueObject)
    fonts = []
    custom_font = settings.get_string('font-family')
    pango_context = self.get_pango_context()
    for font_familie in pango_context.list_families():
      fonts.append(font_familie.get_name())
    fonts.sort(key=str.lower)
    selected = 0
    for index, font in enumerate(fonts):
      model.insert(index, Handy.ValueObject.new(font))
      if font == custom_font:
        selected = index
    self.custom_font_combo.bind_name_model(model, Handy.ValueObject.dup_string)
    self.custom_font_combo.set_selected_index(selected)

  # Populate list of available languages

  def _populate_languages_list(self):
    for lang_id in sorted(languages.wikilangs):
      lang_name = languages.wikilangs[lang_id].capitalize()
      if lang_id in languages.items:
        row = LanguageBoxRow(lang_name, lang_id, True)
      else:
        row = LanguageBoxRow(lang_name, lang_id, False)
      self.languages_list.add(row)
      row.lang_check.connect('toggled', self._language_checkbutton_cb)

  # On window close refresh languages list (if changed)

  def _window_delete_cb(self, prefs_window, event):
    if self._languages_changed:
      languages.clear()
      rows = self.languages_list.get_children()
      for row in rows:
        if row.lang_check.get_active():
          languages.items[row.lang_id] = row.lang_name

      if len(languages.items) == 0:
        languages.items['en'] = 'English'

      window = self.get_transient_for()
      window.headerbar.refresh_langs()

    return False

  # On custom font selected update settings

  def _custom_font_combo_selected_cb(self, custom_font_combo, value):
    model = custom_font_combo.get_model()
    index = custom_font_combo.get_selected_index()
    font = model.get_item(index).dup_string()
    settings.set_string('font-family', font)

  # On button click clear history of visited articles

  def _clear_historic_button_cb(self, clear_historic_button):
    clear_dialog = Gtk.MessageDialog(self,
                                     Gtk.DialogFlags.MODAL,
                                     Gtk.MessageType.WARNING,
                                     Gtk.ButtonsType.CANCEL,
                                     _('Clear Historic?'))

    clear_dialog.set_property('secondary-text', _('The history of visited articles will be deleted permanently.'))
    clear_dialog.add_buttons(_('Clear Historic'), Gtk.ResponseType.YES)

    response = clear_dialog.run()
    clear_dialog.destroy()
    if response == Gtk.ResponseType.CANCEL:
      return

    historic.clear()
    historic.save()

  # On button click clear personal data

  def _clear_data_button_cb(self, clear_data_button):
    clear_dialog = Gtk.MessageDialog(self,
                                     Gtk.DialogFlags.MODAL,
                                     Gtk.MessageType.WARNING,
                                     Gtk.ButtonsType.CANCEL,
                                     _('Clear Personal Data?'))

    clear_dialog.set_property('secondary-text', _('Cookies and cache data will be deleted permanently.'))
    clear_dialog.add_buttons(_('Clear Data'), Gtk.ResponseType.YES)

    response = clear_dialog.run()
    clear_dialog.destroy()
    if response == Gtk.ResponseType.CANCEL:
      return

    web_context = WebKit2.WebContext.get_default()
    data_manager = web_context.get_website_data_manager()
    data_manager.clear(WebKit2.WebsiteDataTypes.ALL, 0, None, None, None)

  # On row selected set language check button

  def _languages_list_selected_cb(self, languages_list, row):
    row.lang_check.set_active(not row.lang_check.get_active())

  # Set languages changed variable on check button changed

  def _language_checkbutton_cb(self, check_button):
    if not self._languages_changed:
      self._languages_changed = True

  # On button click check all languages

  def _select_all_button_cb(self, select_all_button):
    rows = self.languages_list.get_children()
    for row in rows:
      row.lang_check.set_active(True)

  # On button click uncheck all languages

  def _select_none_button_cb(self, select_none_button):
    rows = self.languages_list.get_children()
    for row in rows:
      row.lang_check.set_active(False)


# Class for row in languages list
# Create row from name, id and filter status

@Gtk.Template(resource_path='/com/github/hugolabe/Wike/ui/prefs-row.ui')
class LanguageBoxRow(Gtk.ListBoxRow):

  __gtype_name__ = 'LanguageBoxRow'

  lang_name_label = Gtk.Template.Child()
  lang_id_label = Gtk.Template.Child()
  lang_check = Gtk.Template.Child()

  # Set row values

  def __init__(self, lang_name, lang_id, checked):
    super().__init__()

    self.lang_name = lang_name
    self.lang_id = lang_id
    self.lang_name_label.set_label(lang_name)
    self.lang_id_label.set_label(lang_id)
    self.lang_check.set_active(checked)




# search.py
#
# This file is part of Wike, a Wikipedia Reader for the GNOME Desktop.
# Copyright 2021 Hugo Olabera <hugolabe@gmail.com>.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


from threading import Thread

import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gio, GLib, Gtk

from wike import wikipedia
from wike.data import settings, languages


# Search entry class
# Manage article searchs in Wikipedia

@Gtk.Template(resource_path='/com/github/hugolabe/Wike/ui/search.ui')
class SearchEntry(Gtk.SearchEntry):

  __gtype_name__ = 'SearchEntry'

  # Create suggestions list and connect search entry events

  def __init__(self, search_button, window):
    super().__init__()

    self.window = window

    self.settings_popover = SettingsPopover(self)
    self.suggestions_popover = SuggestionsPopover(self)
    self.results_list = None
    self._results_changed = False

    lang_id = settings.get_string('search-language')
    self.set_icon_tooltip_text(Gtk.EntryIconPosition.PRIMARY, languages.wikilangs[lang_id].capitalize())

    self.connect('show', self._entry_show_cb)
    self.connect('key-press-event', self._key_press_cb, search_button)
    self.connect('icon-release',self._icon_release_cb)

  # Search text in Wikipedia and load results list

  def _search_wikipedia(self, text):
    try:
      self.results_list = wikipedia.search(text.lower(), settings.get_string('search-language'), 10)
    except:
      self.results_list = None
    self._results_changed = True

  # When entry show add timeout function

  def _entry_show_cb(self, entry):
    if self.suggestions_popover.is_visible():
      self.suggestions_popover.hide()

    if settings.get_boolean('search-suggestions'):
      GLib.timeout_add(250, self._timeout_cb)

  # Timeout function which show search results in popover

  def _timeout_cb(self):
    if not self.is_visible():
      return False

    if self._results_changed:
      if self.results_list != None:
        self.suggestions_popover.populate(self.results_list)
        if not self.suggestions_popover.is_visible():
          self.suggestions_popover.show_all()
      else:
        self.suggestions_popover.hide()
      self._results_changed = False
    return True

  # Manage ESC and Down keys

  def _key_press_cb(self, entry, event, search_button):
    if event.keyval == 65364:
      if self.suggestions_popover.is_visible():
        self.suggestions_popover.grab_focus()
      return True
    elif event.keyval == 65307:
      if self.suggestions_popover.is_visible():
        self.suggestions_popover.hide()
      else:
        search_button.set_active(False)

  # Show settings popover on icon clicked

  def _icon_release_cb(self, entry, icon_pos, event):
    if icon_pos == Gtk.EntryIconPosition.PRIMARY:
      if self.suggestions_popover.is_visible():
        self.suggestions_popover.hide()
      self.settings_popover.show_all()

  # When text changes run search in new thread

  def do_search_changed(self):
    if settings.get_boolean('search-suggestions'):
      text = self.get_text()
      if len(text) > 2:
        t = Thread(target=self._search_wikipedia, args=(text, ))
        t.start()
      else:
        self.results_list = None
        self._results_changed = True

  # On entry activated search for text in Wikipedia and load result

  def do_activate(self):
    text = self.get_text()
    if text != '':
      try:
        result = wikipedia.search(text.lower(), settings.get_string('search-language'), 1)
      except:
        self.window.page.wikiview.load_message('notfound', None)
      else:
        if result != None:
          uri = result[1][0]
          self.window.page.wikiview.load_wiki(uri)
        else:
          self.window.page.wikiview.load_message('notfound', None)


# Popover class for search suggestions
# Show search suggestions when user type in entry

class SuggestionsPopover(Gtk.Popover):

  # Create menu and actions for suggestions

  def __init__(self, search_entry):
    super().__init__()
    self.set_relative_to(search_entry)
    self.set_size_request(360, -1)
    self.set_modal(False)

    self._window = search_entry.window

    self._results_menu = Gio.Menu()
    self.bind_model(self._results_menu, None)

    actions = Gio.SimpleActionGroup()
    action = Gio.SimpleAction.new('suggestion', GLib.VariantType('s'))
    action.connect('activate', self._suggestion_activate_cb)
    actions.add_action(action)
    self.insert_action_group('suggestions_popover', actions)

  # Populate search results list

  def populate(self, results_list):
    self._results_menu.remove_all()
    for index, item in enumerate(results_list[0]):
      action_string = 'suggestions_popover.suggestion(\'' + str(index) + '\')'
      button = Gio.MenuItem.new(item, action_string)
      self._results_menu.append_item(button)

  # Set focus to search entry on ESC key press

  def do_key_press_event(self, event):
    if event.keyval == 65307:
      search_entry = self.get_relative_to()
      search_entry.grab_focus_without_selecting()
      self.hide()

  # On menu activated get uri for selected result and load it

  def _suggestion_activate_cb(self, action, parameter):
    index = int(parameter.unpack())
    search_entry = self.get_relative_to()
    uri = search_entry.results_list[1][index]
    self._window.page.wikiview.load_wiki(uri)


# Popover class for search settings
# Select search language and set search suggestions

@Gtk.Template(resource_path='/com/github/hugolabe/Wike/ui/search-settings.ui')
class SettingsPopover(Gtk.Popover):

  __gtype_name__ = 'SettingsPopover'

  filter_entry = Gtk.Template.Child()
  languages_list = Gtk.Template.Child()
  suggestions_switch = Gtk.Template.Child()

  # Populate search languages list and connect signals and bindings

  def __init__(self, search_entry):
    super().__init__()
    self.set_relative_to(search_entry)

    self._window = search_entry.window

    self.languages_list.set_filter_func(self._filter_list)
    self.languages_list.set_header_func(self._set_row_separator)
    self.populate_list()

    settings.bind('search-suggestions', self.suggestions_switch, 'active', Gio.SettingsBindFlags.DEFAULT)
    self.filter_entry.connect('search-changed', self._filter_entry_changed_cb)
    self.languages_list.connect('row-activated', self._languages_list_activated_cb)

  # Filter languages list for filter entry content

  def _filter_list(self, row):
    text = self.filter_entry.get_text()
    if text == '':
      return True

    if row.lang_name.lower().startswith(text.lower()) or row.lang_id.lower().startswith(text.lower()):
      return True
    else:
      return False

  # Set language row separator

  def _set_row_separator(self, row, before):
    if before and not row.get_header():
      separator = Gtk.Separator.new(Gtk.Orientation.HORIZONTAL)
      row.set_header(separator)

  # Populate languages list

  def populate_list(self):
    rows = self.languages_list.get_children()
    for row in rows:
      self.languages_list.remove(row)

    if len(languages.items) > 0:
      for lang_id in sorted(languages.items):
        lang_name = languages.items[lang_id].capitalize()
        check_mark = (lang_id == settings.get_string('search-language'))
        row = SearchLangBoxRow(lang_name, lang_id, check_mark)
        self.languages_list.add(row)

  # Change search language on list activated

  def _languages_list_activated_cb(self, lang_list, row):
    search_entry = self.get_relative_to()
    search_entry.set_icon_tooltip_text(Gtk.EntryIconPosition.PRIMARY, languages.wikilangs[row.lang_id].capitalize())
    settings.set_string('search-language', row.lang_id)
    self.hide()
    message = _('Default search language: ') + row.lang_name
    self._window.show_notification(message)
    self.populate_list()

  # Refresh languages list on filter entry changed

  def _filter_entry_changed_cb(self, filter_entry):
    self.languages_list.invalidate_filter()

  # Unselect items and show popover

  def show_all(self):
    self.languages_list.unselect_all()
    Gtk.Popover.show_all(self)


# Class for row in search languages list
# Create row from language name and id

@Gtk.Template(resource_path='/com/github/hugolabe/Wike/ui/search-row.ui')
class SearchLangBoxRow(Gtk.ListBoxRow):

  __gtype_name__ = 'SearchLangBoxRow'

  label_name = Gtk.Template.Child()
  label_id = Gtk.Template.Child()
  image_check = Gtk.Template.Child()

  # Initialize widgets

  def __init__(self, lang_name, lang_id, check_mark):
    super().__init__()

    self.lang_name = lang_name
    self.lang_id = lang_id

    self.label_name.set_label(lang_name)
    if check_mark:
      self.image_check.set_from_icon_name('emblem-ok-symbolic', Gtk.IconSize.BUTTON)
    self.label_id.set_label(lang_id)

